from ..core_models import ChangeTemplate


class DynamischeExchangeVerteilerAttributes(ChangeTemplate):
    createTimestamp: str | None = None  # Creation date
    displayName: str | None = None  # Displayname
    hideFromAddressLists: str | None = None  # Hide from address lists
    includeInstituteUsersOnly: str | None = None  # Add institute users only
    mail: str | None = None  # Primary email address
    mailboxServer: str | None = None  # Mailbox server
    managedBy: list[str] | None = None  # Managed by
    memberOfStaticExchangeDistGrp: list[str] | None = (
        None  # Member of static distribution group
    )
    modifyTimeStamp: str | None = None  # Modified on
    proxyAddress: list[str] | None = None  # Email addresses
    recipientFilter: str | None = None  # Recipient filter
    sendPermission: list[str] | None = None  # Send permissions
    sendPermissionLevel: str | None = None  # Send permission level



from .benutzerverwaltung import Benutzerverwaltung, Benutzerverwaltung_MDMP  # noqa
from .dynamischeexchangeverteiler import DynamischeExchangeVerteiler  # noqa
from .exchangeverteiler import ExchangeVerteiler  # noqa
from .gruppenverwaltung import Gruppenverwaltung, Gruppenverwaltung_MDMP  # noqa
from .ldapverteiler import LDAPVerteiler  # noqa
from .mailinglisten import Mailinglisten  # noqa
from .ressourcen import Ressourcen  # noqa
from .selfservice import SelfService  # noqa
from .sharedmailbox import SharedMailbox  # noqa



from __future__ import annotations

from pydantic import Field

from ..attributes import GruppenverwaltungAttributes, GruppenverwaltungAttributes_MDMP

try:
    from ..base_models import GruppenverwaltungBase
except ImportError:
    from ..core_models import IDMUserClass

    class GruppenverwaltungBase(IDMUserClass):
        pass


from ..core_models import ChangeTemplate


class Gruppenverwaltung(GruppenverwaltungBase):
    api_suffix: str = Field(default="/Gruppenverwaltung/objects")
    user_class: type[ChangeTemplate] = Field(default=GruppenverwaltungAttributes)


class Gruppenverwaltung_MDMP(GruppenverwaltungBase):
    api_suffix: str = Field(default="/Gruppenverwaltung/objects")
    user_class: type[ChangeTemplate] = Field(default=GruppenverwaltungAttributes_MDMP)



from ..core_models import ChangeTemplate


class GruppenverwaltungAttributes(ChangeTemplate):
    awsRole: list[str] | None = None  # AWS role mapping
    cn: str | None = None  # GOESTERN-ID
    description: str | None = None  # Description
    displayName: str | None = None  # Displayname
    excludeMember: list[str] | None = None  # Excluded members
    goesternExpirationDate: str | None = None  # Expiration date
    goesternSamAccountName: str | None = None  # Account name in source system
    groupMemberFilter: str | None = None  # Group member filter
    groupMemberFilterAttribute: str | None = (
        None  # Attribute name for determining members
    )
    groupMemberFilterError: str | None = None  # Group calculation error
    groupMemberLimit: str | None = None  # group member limit
    groupMemberScope: str | None = None  # Group member scope
    hideFromAddressLists: str | None = None  # Hide from address lists
    includeMember: list[str] | None = None  # Additional members
    isProtected: str | None = None  # Only editable by "Managed by"
    mail: str | None = None  # Primary email address
    mailboxServer: str | None = None  # Mailbox server
    managedBy: list[str] | None = None  # Managed by
    member: list[str] | None = None  # Members
    nameSpace: str | None = None  # Namespace
    ou: str | None = None  # Institute
    proxyAddress: list[str] | None = None  # Email addresses
    scope: list[str] | None = None  # Visibility
    sendPermission: list[str] | None = None  # Send permissions
    sendPermissionLevel: str | None = None  # Send permission level


class GruppenverwaltungAttributes_MDMP(GruppenverwaltungAttributes):
    dn: str | None = None  # DN



from ..core_models import ChangeTemplate


class SelfServiceAttributes(ChangeTemplate):
    cn: str | None = None  # GOESTERN-ID
    effectivePrivilege: list[str] | None = None  # Privileges
    givenName: str | None = None  # First name
    goesternProxyAddresses: list[str] | None = None  # Email addresses
    mail: str | None = None  # Primary email address
    ou: str | None = None  # Institute
    sn: str | None = None  # Last name
    uid: str | None = None  # Username
    umgArztnummer: str | None = None  # Doctor number



import pydantic

from ..core_models import ChangeTemplate


class BenutzerverwaltungAttributes(ChangeTemplate):
    accountType: str | None = None  # Account type
    additionalPasswordExpirationTime: str | None = (
        None  # Password expiration date (additional password)
    )
    additionalPasswordModifyTime: str | None = (
        None  # Last password change (additional password)
    )
    city: str | None = None  # City
    createTimestamp: str | None = None  # Creation date
    departmentNumber: str | None = None  # Department
    description: str | None = None  # Description
    eduPersonPrincipalName: str | None = None  # eduPersonPrincipalName
    eduPersonScopedAffiliation: list[str] | None = None  # eduPersonScopedAffiliation
    effectivePrivilege: list[str] | None = None  # Privileges
    employeeNumber: str | None = None  # Employee number
    employeeStatus: str | None = None  # Employee status
    employeeType: str | None = None  # Employee type
    exchangeTargetAddress: str | None = None  # Exchange redirect address
    externalEmailAddress: list[str] | None = None  # External email addresses
    facsimileTelephoneNumber: str | None = None  # Fax number
    filterAttribute1: list[str] | None = None  # Filter attribute 1
    filterAttribute2: list[str] | None = None  # Filter attribute 2
    filterAttribute3: list[str] | None = None  # Filter attribute 3
    givenName: str | None = None  # First name
    goesternDisableDate: str | None = None  # Disable/Enable date
    goesternDisableReason: str | None = None  # Disable reason
    goesternExchHideFromAddressLists: str | None = None  # Hide from address lists
    goesternExchangeQuota: str | None = None  # Exchange quota
    goesternExpirationDate: str | None = None  # Expiration date
    goesternGWDGadDisplayName: str | None = None  # AD-Displayname
    goesternLockoutTime: str | None = None  # Short time lockout (AD)
    goesternMailRoutingAddresses: list[str] | None = None  # Routing addresses
    goesternMailboxServer: str | None = None  # Mailbox server
    goesternMailboxZugehoerigkeit: str | None = None  # Mailbox type
    goesternProxyAddresses: list[str] | None = None  # Email addresses
    goesternRemoveDate: str | None = None  # Remove date
    goesternSAMAccountName: str | None = None  # Account name in source system
    goesternUserStatus: str | None = None  # User status
    goesternUserType: str | None = None  # User type
    isEnforceMFA: str | None = None  # Multi-factor authentication activated
    isEnforceSSOFederatedLogin: str | None = None  # Federated account
    isInitialAdditionalPassword: str | None = (
        None  # Additional password is initial password
    )
    isInitialPassword: str | None = None  # Initial password
    isScheduledForDeletion: str | None = None  # Deferred deletion
    l: str | None = None  # Country
    loginExpirationTime: str | None = None  # Login expiration date
    mail: str | None = None  # Primary email address
    memberOfStaticExchangeDistGrp: list[str] | None = None  # Group memberships
    mobile: str | None = None  # Mobile
    modifyTimeStamp: str | None = None  # Modified on
    mpgEmployeeNumber: str | None = None  # Employee number MPG
    mwsEmployeeNumber: str | None = None  # MWS employee number
    oldUid: list[str] | None = None  # Previous usernames
    ou: str | None = None  # Institute
    ownCloudQuota: str | None = None  # OwnCloud Quota
    passwordExpirationTime: str | None = None  # Password expiration date
    postalCode: str | None = None  # Postal code
    pwdChangedTime: str | None = None  # Last password change
    responsiblePerson: list[str] | None = None  # Responsible user
    roomNumber: str | None = None  # Room number
    sn: str | None = None  # Last name
    st: str | None = None  # State
    street: str | None = None  # Street
    telephoneNumber: str | None = None  # Telefonnumber
    title: str | None = None  # Job title
    uid: str | None = None  # Username
    voPersonExternalAffiliation: list[str] | None = (
        None  # voPerson Scoped External Affiliation
    )

    @pydantic.field_validator("isScheduledForDeletion")
    @classmethod
    def isScheduledForDeletion_validator(cls, value):
        if value is None:
            return value
        return value.upper()


class BenutzerverwaltungAttributes_MDMP(BenutzerverwaltungAttributes):
    dn: str | None = None  # DN
    externalTaskCommand: str | None = None  # External Task command
    throttlingpolicy: str | None = None  # Throttling Policy
    workGroup: str | None = None  # WorkGroup



import datetime
import os
import uuid

from gwdg_idm_api import Benutzerverwaltung_MDMP as Benutzerverwaltung
from gwdg_idm_api import CreateBenutzerverwaltung_MDMP as NewGWDGUser
from gwdg_idm_api.util import AlreadyDeletedError, pretty_print

# Load idm credentials
# Do not upload credentials in any way to github/gitlab etc.
idm_username = os.environ["IDM_USERNAME"]
idm_password = os.environ["IDM_PASSWORD"]


# For testing purposes, we are using the idm-stage endpoint.
# This one is a copy of the previous day and can be used
# for sandbox testing.
# Change to "https://idm.gwdg.de/api/v3" for production.
if "stage" not in os.environ["IDM_API_URL"]:
    print("There is no stage in IDM_API_URL! This is not supported!")
    exit(1)
benutzerverwaltung = Benutzerverwaltung(
    api_url=os.environ["IDM_API_URL"],
    username=idm_username,
    password=idm_password,
)

# Get all users having their surename starting with 'a'
sn_a_users = benutzerverwaltung.get_multiple("$sn -eq 'a*'")
print("# First two entries starting with 'a'")
print(pretty_print(sn_a_users[:2]))

# Take the first user and print a more detailed view
print("# Get detailed information")
single_user = benutzerverwaltung.get_single(sn_a_users[0].id)
print(pretty_print(single_user))

single_user = benutzerverwaltung.create_new(
    NewGWDGUser(
        givenName=str(uuid.uuid4())[:8],
        sn=str(uuid.uuid4())[:8],
        workGroup="best_group",
        departmentNumber="best_department",
    )
)
print("# Create new user")
print(pretty_print(single_user))

# Change the lastname of the user to 'Upps bad lastnäme' and workGroup to '-'
# and check if the update happened
print("# Update user")
updated_user = benutzerverwaltung.update_single(
    single_user, {"workGroup": "-", "sn": "Upps bad lastnäme"}
)
print(pretty_print(updated_user))

# Ah that did not work out well!
# Now lets remove the user in 10 days (set goesternExpirationDate)
# If
print("# Delete user in 10 days")
try:
    deleted_user = benutzerverwaltung.delete_user(
        single_user, datetime.datetime.now() + datetime.timedelta(days=10)
    )
except AlreadyDeletedError:
    # If the user is already deleted an AlreadyDeletedError is thrown
    print("Already deleted")
else:
    print(pretty_print(deleted_user))

# Actually, lets delete him yesterday
# The delete_user functions will automatically just check the
# isScheduledForDeletion flag if the expire_datetime is in the past
print("# Delete user now")
try:
    deleted_user = benutzerverwaltung.delete_user(
        updated_user, datetime.datetime.now() - datetime.timedelta(days=1)
    )
except AlreadyDeletedError:
    print("Already deleted")
else:
    print(pretty_print(deleted_user))

# Lets create a new user
print("# Create new user")
new_user_tmp = NewGWDGUser(
    givenName="john", sn="doe", workGroup="-", departmentNumber="ZE"
)
new_user = benutzerverwaltung.create_new(new_user_tmp)
print(pretty_print(new_user))

print("# All done :)")



from __future__ import annotations

import datetime
import logging
import typing

from .core_models import ChangeTemplate, IDMRequest, IDMUserClass
from .util import AlreadyDeletedError, IDMRequestError

if typing.TYPE_CHECKING:
    from requests import Response

    from .core_models import PasswordTemplate


class Workspaces(IDMRequest):
    api_suffix: str = "/workspaces"

    def get_workspaces_json(self):
        return self.get_suffix()

    def get_workspace_attributes(self):
        json_resp = self.get_workspaces_json()
        workspace_list = json_resp["workspaceList"]
        workspaces = {}
        for entry in workspace_list:
            workspaces[entry["name"]] = entry["attributes"]
        return workspaces


class BenutzerverwaltungBase(IDMUserClass):
    def delete_user(
        self, user: ChangeTemplate, expire_datetime: datetime.datetime | datetime.date
    ):
        today: datetime.date
        expire_date: datetime.date
        activate_now: bool
        change_data: dict[str, str]

        today = datetime.datetime.now().date()
        try:
            expire_date = expire_datetime.date()
        except AttributeError:
            expire_date = expire_datetime

        current_user_status = self.get_single(user.id)

        activate_now = (expire_date - today).total_seconds() <= 0
        status_deleted = current_user_status.goesternUserStatus in ("2", "255")

        change_data = {}
        change_data["isScheduledForDeletion"] = "FALSE"
        change_data["goesternExpirationDate"] = ""

        self.update_single(user, change_data, single_call=True)  # type: ignore

        if activate_now and not status_deleted:
            change_data = {
                "isScheduledForDeletion": "TRUE",
            }
        elif activate_now:
            change_data = {}
        else:
            # TODO: After fixing the 'year 2038' problem in the IDM, this if statement can be removed again.
            if expire_date.year > 2037:
                logging.warn(
                    "Expire date year is bigger than 2037. Reduce expire date to 2038."
                )
                expire_date = datetime.date(2038, 1, 1)
            change_data = {
                "goesternExpirationDate": expire_date.strftime("%d.%m.%Y"),
            }

        return self.update_single(user, change_data)  # type: ignore

    def reactivate_user(self, user: ChangeTemplate, create_mailbox: bool = True):
        current_user_status = self.get_single(user.id)

        if current_user_status.goesternUserStatus == "255":
            raise AlreadyDeletedError
        elif current_user_status.goesternUserStatus != "2":
            return user

        if current_user_status.goesternExpirationDate:
            change_data = {"goesternExpirationDate": ""}
            user = self.update_single(user, change_data)  # type: ignore

        change_data = {
            "goesternUserStatus": "0",
        }
        if create_mailbox:
            change_data["externalTaskCommand"] = "CreateMailbox"
        return self.update_single(user, change_data)  # type: ignore

    def change_password(self, user: PasswordTemplate | list[PasswordTemplate]) -> list:
        resp: Response
        data: str

        if not isinstance(user, list):
            user = [user]

        pdfs = []
        for cur_user in user:
            request_url: str = (
                f"{self.api_url}{self.api_suffix}/{cur_user.id}/{cur_user.template_name}{cur_user.pdf_suffix}"
            )
            data = cur_user.to_json()
            resp = self.post_request(request_url, data)
            if cur_user.pdf_suffix:
                pdfs.append(resp.content)

            else:
                if "success" not in resp.text:
                    raise IDMRequestError(
                        f"Could not connect to IDM: Invalid combination\n{resp.text=}"
                    )
        return pdfs


class GruppenverwaltungBase(IDMUserClass):
    def delete_user(self, user: ChangeTemplate, remark: str = ""):
        request_url: str = (
            f"{self.api_url}/Deletion/DeleteObject{self.api_suffix}/{user.id}"
        )
        request_url = f"{self.api_url}{self.api_suffix}/objects/{user.id}/delete"
        data = {"_remark": remark}
        print(request_url)
        resp = self.post_request(request_url, data)
        print(resp.text)



from __future__ import annotations

from pydantic import Field

from ..attributes import BenutzerverwaltungAttributes, BenutzerverwaltungAttributes_MDMP

try:
    from ..base_models import BenutzerverwaltungBase
except ImportError:
    from ..core_models import IDMUserClass

    class BenutzerverwaltungBase(IDMUserClass):
        pass


from ..core_models import ChangeTemplate


class Benutzerverwaltung(BenutzerverwaltungBase):
    api_suffix: str = Field(default="/Benutzerverwaltung/objects")
    user_class: type[ChangeTemplate] = Field(default=BenutzerverwaltungAttributes)


class Benutzerverwaltung_MDMP(BenutzerverwaltungBase):
    api_suffix: str = Field(default="/Benutzerverwaltung/objects")
    user_class: type[ChangeTemplate] = Field(default=BenutzerverwaltungAttributes_MDMP)



from __future__ import annotations

from pydantic import Field

from ..attributes import DynamischeExchangeVerteilerAttributes

try:
    from ..base_models import DynamischeExchangeVerteilerBase
except ImportError:
    from ..core_models import IDMUserClass

    class DynamischeExchangeVerteilerBase(IDMUserClass):
        pass


from ..core_models import ChangeTemplate


class DynamischeExchangeVerteiler(DynamischeExchangeVerteilerBase):
    api_suffix: str = Field(default="/DynamischeExchangeVerteiler/objects")
    user_class: type[ChangeTemplate] = Field(
        default=DynamischeExchangeVerteilerAttributes
    )



from ..core_models import ChangeTemplate


class LDAPVerteilerAttributes(ChangeTemplate):
    createTimestamp: str | None = None  # Creation date
    description: str | None = None  # Description
    filterAttribute1: list[str] | None = None  # Filter attribute 1
    filterAttribute2: list[str] | None = None  # Filter attribute 2
    filterAttribute3: list[str] | None = None  # Filter attribute 3
    goesternExchHideFromAddressLists: str | None = None  # Hide from address lists
    goesternExpirationDate: str | None = None  # Expiration date
    goesternGWDGadDisplayName: str | None = None  # Displayname
    goesternMailroutingAddresses: list[str] | None = None  # Routing addresses
    goesternProxyAddresses: list[str] | None = None  # Email addresses
    goesternUserType: str | None = None  # User type
    mail: str | None = None  # Primary email address
    managedBy: list[str] | None = None  # Managed by
    modifyTimeStamp: str | None = None  # Modified on
    uid: str | None = None  # Username



from ..core_models import ChangeTemplate


class RessourcenAttributes(ChangeTemplate):
    capacity: str | None = None  # Capacity
    city: str | None = None  # City
    createTimestamp: str | None = None  # Creation date
    delegate: list[str] | None = None  # Delegate
    department: str | None = None  # Department
    description: str | None = None  # Description
    displayName: str | None = None  # Displayname
    fullAccessAllowed: list[str] | None = None  # Full access
    hideFromAddressLists: str | None = None  # Hide from address lists
    mail: str | None = None  # Primary email address
    mailboxServer: str | None = None  # Mailbox server
    managedBy: list[str] | None = None  # Managed by
    memberOfStaticExchangeDistGrp: list[str] | None = (
        None  # Member of static distribution group
    )
    modifyTimeStamp: str | None = None  # Modified on
    ou: str | None = None  # Institute
    postalCode: str | None = None  # Postal code
    proxyAddress: list[str] | None = None  # Email addresses
    resourceType: str | None = None  # Resource type
    roomNumber: str | None = None  # Room number
    st: str | None = None  # State
    street: str | None = None  # Street
    telephoneNumber: str | None = None  # Telefonnumber



import argparse
import inspect
import os
import pathlib
import re
from dataclasses import dataclass

from gwdg_idm_api import Workspaces


@dataclass
class Config:
    output_folder: pathlib.Path
    username: str
    password: str
    api_url: str

    _attribute_folder: pathlib.Path | None = None
    _model_folder: pathlib.Path | None = None
    _create_templates_folder: pathlib.Path | None = None

    @property
    def attribute_folder(self):
        if self._attribute_folder is None:
            folder = self.output_folder / "attributes"
            folder.mkdir(exist_ok=True, parents=True)
            self._attribute_folder = folder
        return self._attribute_folder

    @property
    def models_folder(self):
        if self._model_folder is None:
            folder = self.output_folder / "models"
            folder.mkdir(exist_ok=True, parents=True)
            self._model_folder = folder
        return self._model_folder

    @property
    def create_templates_folder(self):
        if self._create_templates_folder is None:
            folder = self.output_folder / "create_templates"
            folder.mkdir(exist_ok=True, parents=True)
            self._create_templates_folder = folder
        return self._create_templates_folder


def args_to_config(args) -> Config:
    return Config(
        pathlib.Path(args.output_folder),
        os.environ["IDM_USERNAME"],
        os.environ["IDM_PASSWORD"],
        os.environ["IDM_API_URL"],
    )


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("output_folder")
    return parser.parse_args()


def get_workspaces(config: Config):
    check = Workspaces(
        username=config.username, password=config.password, api_url=config.api_url
    )
    return check.get_workspace_attributes()


def create_attributes(workspaces: dict, config: Config):
    add_attributes = {
        "benutzerverwaltung": {
            "MDMP": [
                {"name": "dn", "multiValue": False, "displayName": "DN"},
                {"name": "workGroup", "multiValue": False, "displayName": "WorkGroup"},
                {
                    "name": "throttlingpolicy",
                    "multiValue": False,
                    "displayName": "Throttling Policy",
                },
                {
                    "name": "externalTaskCommand",
                    "multiValue": False,
                    "displayName": "External Task command",
                },
            ]
        },
        "gruppenverwaltung": {
            "MDMP": [{"name": "dn", "multiValue": False, "displayName": "DN"}]
        },
    }

    ### Helper functions to be inserted as validators!
    def to_upper(cls, value):
        if value is None:
            return value
        return value.upper()

    ### Helper functions end

    validators = {
        "benutzerverwaltung": {
            "isScheduledForDeletion": to_upper,
        }
    }

    init_string = []
    created_classes = {}
    for key, attributes in workspaces.items():
        class_names = {}
        base_class = f"{key}Attributes"
        class_names[""] = base_class

        out_string = []
        out_string.append("from ..core_models import ChangeTemplate")
        out_string.append("")
        out_string.append("")

        cur_validator = validators.get(key.lower(), {})
        class_string = create_attribute_class(
            base_class, "ChangeTemplate", attributes, cur_validator
        )
        out_string.extend(class_string)

        for new_key, new_attrs in add_attributes.get(key.lower(), {}).items():
            new_base_class = f"{base_class}_{new_key}"
            class_names[new_key] = new_base_class
            class_string = create_attribute_class(
                new_base_class, base_class, new_attrs, cur_validator
            )
            out_string.append("")
            out_string.append("")
            out_string.extend(class_string)

        out_string.append("")
        out_text = "\n".join(out_string)

        if "pydantic." in out_text:
            out_text = "import pydantic\n" + out_text

        (config.attribute_folder / f"{key.lower()}.py").write_text(out_text)

        init_string.append(
            f"from .{key.lower()} import {', '.join(sorted(class_names.values()))}  # noqa"
        )

        created_classes[key] = class_names

    (config.attribute_folder / "__init__.py").write_text("\n".join(sorted(init_string)))

    return created_classes


def create_attribute_class(class_name, inherit_name, attributes, validators):
    out_string = []
    out_string.append(f"class {class_name}({inherit_name}):")
    inner_strings = []
    old_attrs = set()
    for attr in attributes:
        assert attr["name"] not in old_attrs, (class_name, attr["name"], old_attrs)
        old_attrs.add(attr["name"])
        inner_string = []
        inner_string.append(f"    {attr['name']}:")
        if attr["multiValue"]:
            inner_string.append("list[str]")
        else:
            inner_string.append("str")
        if "mandatory" not in attr or not attr["mandatory"]:
            inner_string.append("|")
            inner_string.append("None")
            inner_string.append("=")
            inner_string.append(
                f'"{attr["default"]}"'
                if "default" in attr and attr["default"] is not None
                else "None"
            )
        if "mandatory" in attr and attr["mandatory"] and "default" in attr:
            inner_string.append("=")
            inner_string.append(f'"{attr["default"]}"')
        inner_string.append(" #")
        inner_string.append(attr["displayName"])
        inner_strings.append(" ".join(inner_string))
    out_string.extend(sorted(inner_strings))
    for vali_key, vali_value in validators.items():
        if vali_key in old_attrs:
            out_string.append("")
            out_string.append(f'    @pydantic.field_validator("{vali_key}")')
            out_string.append("    @classmethod")
            function_source = "".join(inspect.getsource(vali_value))
            function_source = re.sub(
                "def [^(]+", f"def {vali_key}_validator", function_source
            )
            out_string.append(function_source)
    return out_string


def create_models(created_classes, config):

    init_string = []
    for base_name, class_dict in created_classes.items():
        out_string = []
        out_string.append("from __future__ import annotations")
        out_string.append("")
        out_string.append("from pydantic import Field")
        out_string.append("")
        out_string.append(
            f"from ..attributes import {', '.join(sorted(class_dict.values()))}"
        )
        out_string.append("")
        out_string.append(
            f"""
try:
    from ..base_models import {base_name}Base
except ImportError:
    from ..core_models import IDMUserClass

    class {base_name}Base(IDMUserClass):
        pass
            """.strip()
        )
        out_string.append("")
        out_string.append("from ..core_models import ChangeTemplate")
        class_names = []
        for prefix, class_name in class_dict.items():
            if prefix:
                suffix = f"_{prefix}"
            else:
                suffix = ""
            new_class_name = f"{base_name}{suffix}"
            class_names.append(new_class_name)
            out_string.append("")
            out_string.append("")
            out_string.append(
                f"""
class {new_class_name}({base_name}Base):
    api_suffix: str = Field(default="/{base_name}/objects")
    user_class: type[ChangeTemplate] = Field(default={class_name})
            """.strip()
            )

        out_string.append("")
        (config.models_folder / f"{base_name.lower()}.py").write_text(
            "\n".join(out_string)
        )

        init_string.append(
            f"from .{base_name.lower()} import {', '.join(sorted(class_names))}  # noqa"
        )

    (config.models_folder / "__init__.py").write_text("\n".join(sorted(init_string)))


def create_templates(workspaces: dict, config: Config):
    attributes = {
        "benutzerverwaltung": {
            "MDMP": [
                {
                    "name": "create_template_name",
                    "multiValue": False,
                    "displayName": "Name of the template",
                    "default": "MDMP",
                    "mandatory": True,
                },
                {
                    "name": "givenName",
                    "multiValue": False,
                    "displayName": "first Name",
                    "mandatory": True,
                },
                {
                    "name": "sn",
                    "multiValue": False,
                    "displayName": "last Name",
                    "mandatory": True,
                },
                {
                    "name": "workGroup",
                    "multiValue": False,
                    "displayName": "Workgroup",
                    "mandatory": True,
                },
                {
                    "name": "departmentNumber",
                    "multiValue": False,
                    "displayName": "Department name",
                    "mandatory": True,
                },
                {
                    "name": "roomNumber",
                    "multiValue": False,
                    "displayName": "Office room number",
                    "default": "",
                },
                {
                    "name": "telephoneNumber",
                    "multiValue": False,
                    "displayName": "telephoneNumber",
                    "default": "",
                },
                {
                    "name": "goesternUserType",
                    "multiValue": False,
                    "displayName": "Type of user! 0 normal, 1 function",
                    "default": "0",
                },
                {
                    "name": "externalTaskCommand",
                    "multiValue": False,
                    "displayName": "Task to perform on submission",
                    "default": "CreateMailbox",
                },
            ]
        },
        "gruppenverwaltung": {
            "MDMP": [
                {
                    "name": "create_template_name",
                    "multiValue": False,
                    "displayName": "Name of the template",
                    "default": "MDMP",
                    "mandatory": True,
                },
                {
                    "name": "goesternSamAccountName",
                    "multiValue": False,
                    "displayName": "actual group name",
                    "mandatory": True,
                },
                {
                    "name": "displayName",
                    "multiValue": False,
                    "displayName": "Display name",
                    "mandatory": True,
                },
                {
                    "name": "nameSpace",
                    "multiValue": False,
                    "displayName": "Namespace",
                    "mandatory": True,
                    "default": "MPG",
                },
                {
                    "name": "description",
                    "multiValue": False,
                    "displayName": "Description",
                    "mandatory": False,
                },
                {
                    "name": "goesternExpirationDate",
                    "multiValue": False,
                    "displayName": "Expiration date",
                    "mandatory": False,
                },
                {
                    "name": "member",
                    "multiValue": True,
                    "displayName": "member by internal ID",
                    "mandatory": False,
                },
                {
                    "name": "managedBy",
                    "multiValue": True,
                    "displayName": "Managed by",
                    "mandatory": False,
                },
                {
                    "name": "isProtected",
                    "multiValue": False,
                    "displayName": "Make protected group",
                    "mandatory": False,
                },
                {
                    "name": "proxyAddress",
                    "multiValue": True,
                    "displayName": "Email addresses",
                    "mandatory": False,
                },
                {
                    "name": "mail",
                    "multiValue": False,
                    "displayName": "Primary mail address",
                    "mandatory": False,
                },
                {
                    "name": "hideFromAddressLists",
                    "multiValue": False,
                    "displayName": "Hide from address list",
                    "mandatory": False,
                },
                {
                    "name": "sendPermission",
                    "multiValue": True,
                    "displayName": "Senders allowed to send emails",
                    "mandatory": False,
                },
                {
                    "name": "sendPermissionLevel",
                    "multiValue": False,
                    "displayName": "Send permission level",
                    "mandatory": False,
                },
            ],
        },
    }

    validators = {}

    init_string = []
    for key in workspaces:
        out_string = []
        out_string.append("from ..core_models import CreateTemplate")

        cur_validator = validators.get(key.lower(), {})
        class_names = []
        try:
            cur_attributes = attributes[key.lower()]
        except KeyError:
            continue
        for new_key, new_attrs in cur_attributes.items():
            class_name = f"Create{key}_{new_key}"
            class_names.append(class_name)

            class_string = create_attribute_class(
                class_name, "CreateTemplate", new_attrs, cur_validator
            )
            out_string.append("")
            out_string.append("")
            out_string.extend(class_string)

        out_string.append("")
        out_text = "\n".join(out_string)

        if "pydantic." in out_text:
            out_text = "import pydantic\n" + out_text

        (config.create_templates_folder / f"{key.lower()}.py").write_text(out_text)

        init_string.append(
            f"from .{key.lower()} import {', '.join(sorted(class_names))}  # noqa"
        )

    (config.create_templates_folder / "__init__.py").write_text(
        "\n".join(sorted(init_string))
    )


def main(config: Config):
    workspaces = get_workspaces(config)

    # Attributes
    created_classes = create_attributes(workspaces, config)

    # Models
    create_models(created_classes, config)

    # Create templates
    create_templates(workspaces, config)


if __name__ == "__main__":
    args = parse_args()
    config = args_to_config(args)
    main(config)



from ..core_models import CreateTemplate


class CreateGruppenverwaltung_MDMP(CreateTemplate):
    create_template_name: str = "MDMP"  # Name of the template
    description: str | None = None  # Description
    displayName: str  # Display name
    goesternExpirationDate: str | None = None  # Expiration date
    goesternSamAccountName: str  # actual group name
    hideFromAddressLists: str | None = None  # Hide from address list
    isProtected: str | None = None  # Make protected group
    mail: str | None = None  # Primary mail address
    managedBy: list[str] | None = None  # Managed by
    member: list[str] | None = None  # member by internal ID
    nameSpace: str = "MPG"  # Namespace
    proxyAddress: list[str] | None = None  # Email addresses
    sendPermission: list[str] | None = None  # Senders allowed to send emails
    sendPermissionLevel: str | None = None  # Send permission level



from .attributes import *  # noqa
from .base_models import Workspaces  # noqa
from .core_models import ChangeTemplate, CreateTemplate  # noqa
from .create_templates import *  # noqa
from .models import *  # noqa



from __future__ import annotations

from pydantic import Field

from ..attributes import RessourcenAttributes

try:
    from ..base_models import RessourcenBase
except ImportError:
    from ..core_models import IDMUserClass

    class RessourcenBase(IDMUserClass):
        pass


from ..core_models import ChangeTemplate


class Ressourcen(RessourcenBase):
    api_suffix: str = Field(default="/Ressourcen/objects")
    user_class: type[ChangeTemplate] = Field(default=RessourcenAttributes)



from ..core_models import CreateTemplate


class CreateBenutzerverwaltung_MDMP(CreateTemplate):
    create_template_name: str = "MDMP"  # Name of the template
    departmentNumber: str  # Department name
    externalTaskCommand: str | None = "CreateMailbox"  # Task to perform on submission
    givenName: str  # first Name
    goesternUserType: str | None = "0"  # Type of user! 0 normal, 1 function
    roomNumber: str | None = ""  # Office room number
    sn: str  # last Name
    telephoneNumber: str | None = ""  # telephoneNumber
    workGroup: str  # Workgroup



from __future__ import annotations

import json
import logging
import os
import typing
import urllib.parse

import pydantic
import requests

from .util import (
    BadJsonError,
    IDMNotReachableError,
    IDMRequestError,
    UnexpectedJsonError,
)

if typing.TYPE_CHECKING:
    from requests import Response


__all__ = ["ChangeTemplate", "PasswordTemplate"]


DATA_TYPE = str | list | dict | None


class IDMRequest(pydantic.BaseModel):
    username: str
    password: str
    api_url: str = os.environ.get("IDM_API_URL", "")
    api_suffix: str = ""

    def run_request(
        self, func: typing.Callable[..., Response], url: str, data: DATA_TYPE = None
    ) -> Response:
        try:
            return func(
                url,
                auth=(self.username, self.password),
                timeout=240,  # Needs to be this high to allow for an all-objects query
                headers={
                    "Accept": "application/json",
                    "Accept-Language": "en",
                    "Content-Type": "application/json",
                },
                data=data,
            )
        except Exception as e:
            raise IDMNotReachableError(
                f"Could not connect to IDM: IDM not reachable!\n{e}"
            )

    def get_request(self, url: str) -> Response:
        return self.run_request(requests.get, url)

    def put_request(self, url: str, data: DATA_TYPE) -> Response:
        return self.run_request(requests.put, url, data)

    def post_request(self, url: str, data: DATA_TYPE) -> Response:
        return self.run_request(requests.post, url, data)

    def get_suffix(self, filter_string: str | None = None) -> list[object] | dict:
        request_url = f"{self.api_url}{self.api_suffix}"
        if filter_string is not None:
            request_url = f"{request_url}?filter={urllib.parse.quote(filter_string)}"
        resp = self.get_request(request_url)
        try:
            return json.loads(resp.text)
        except json.decoder.JSONDecodeError:
            raise BadJsonError(resp.text)


class PasswordTemplate(pydantic.BaseModel):
    template_name: str = "changepassword"
    id: str  # GWDG idm ID
    password: str
    isinitial: bool = True
    create_pdf: bool = False

    @property
    def pdf_suffix(self) -> str:
        if self.create_pdf:
            return "/pdf"
        else:
            return ""

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    def to_dict(self) -> dict:
        data = [
            {
                "name": "password",
                "value": [self.password],
            }
        ]
        if self.isinitial:
            data.append(
                {
                    "name": "isinitial",
                    "value": ["true"],
                }
            )
        return {"attributes": data}


class ChangeTemplate(pydantic.BaseModel):
    id: str

    @staticmethod
    def update_dict(name: str, value: str | list[str] | None) -> dict:
        if value is not None:
            value = [value] if not isinstance(value, list) else value
        else:
            value = []
        data = {
            "name": name,
            "value": value,
        }
        return {"attributes": [data]}

    @staticmethod
    def update_json(name: str, value: str | list[str] | None) -> str:
        return json.dumps(ChangeTemplate.update_dict(name, value))

    @classmethod
    def from_json(cls, json: dict) -> ChangeTemplate:
        response_dict: dict[str, list[str] | str]

        response_dict = {
            "id": [json["id"]],
            "dn": [json["dn"]],
        }
        response_dict.update(
            {entry["name"]: entry["value"] for entry in json["attributes"]}
        )

        remove_keys = []
        for key, value in sorted(response_dict.items()):
            try:
                ann: type
                try:
                    ann = cls.model_fields[key].annotation.__args__[0]
                except AttributeError:
                    ann = cls.model_fields[key].annotation

                try:
                    outer_type = ann.__origin__
                except Exception:
                    outer_type = ann

                expected_type = type(outer_type())
            except KeyError:
                remove_keys.append(key)
                logging.warning(
                    "\n"
                    "  key unknown to model\n"
                    f"  User: {response_dict['id']}\n"
                    f"  Key: {key}"
                    f"  Value: {value}"
                )
                continue
            if isinstance(value, expected_type):
                continue

            if isinstance(value, expected_type):
                continue
            elif expected_type is str and isinstance(value, list):
                if len(value) > 1:
                    logging.warning(
                        "\n"
                        "  str expected, but found list: Using first element\n"
                        "  Please check your class specifications.\n"
                        f"  User: {response_dict['id']}\n"
                        f"  Key: {key}"
                        f"  Value: {value}"
                    )
                try:
                    new_val: str = value[0]
                except IndexError:
                    logging.warning(
                        "  str expected, but empty list found: Set to empty string\n"
                        "  Please check your class specifications.\n"
                        f"  User: {response_dict['id']}\n"
                        f"  Key: {key}"
                        f"  Value: {value}"
                    )
                    new_val: str = ""
                response_dict[key] = new_val
            else:
                assert False, (
                    "  Only str and list types are supported so far!"
                    "  Please check your class specifications.\n"
                    f"  User: {response_dict['id']}\n"
                    f"  Key: {key}"
                    f"  Value: {value}"
                )

        return cls(**{key: value for key, value in response_dict.items() if key not in set(remove_keys)})  # type: ignore


class CreateTemplate(pydantic.BaseModel):
    create_template_name: str

    def to_dict(self) -> dict:
        data = [
            {
                "name": key.removeprefix("_"),
                "value": [value] if not isinstance(value, list) else value,
            }
            for key, value in self.dict().items()
            if key != "create_template_name"
        ]
        return {"attributes": data}

    def to_json(self) -> str:
        return json.dumps(self.to_dict())


class IDMUserClass(IDMRequest):
    user_class: type[ChangeTemplate]

    def set_user_class(self, user_class: type[ChangeTemplate]):
        self.user_class = user_class

    def get_single(self, object_id: str) -> ChangeTemplate:
        request_url = f"{self.api_url}{self.api_suffix}/{object_id}"
        resp = self.get_request(request_url)
        try:
            json_resp = json.loads(resp.text)
        except json.decoder.JSONDecodeError:
            raise BadJsonError(resp.text)

        return self.user_class.from_json(json_resp)

    def get_multiple(self, filter_string: str | None = None) -> list[ChangeTemplate]:
        objects = self.get_suffix()
        try:
            return [self.user_class.from_json(obj) for obj in objects["Objects"]]
        except Exception:
            raise UnexpectedJsonError(objects)

    def update_single(
        self,
        user: ChangeTemplate,
        update_dict: dict[str, str | list[str] | None],
        *,
        single_call: bool = False,
    ) -> ChangeTemplate:
        request_url: str = f"{self.api_url}{self.api_suffix}/{user.id}"

        if "mail" in update_dict:
            # Run mail at the very last to avoid missing proxyMail entries
            mail = update_dict.pop("mail")
            update_dict["mail"] = mail

        if not single_call:
            for key, value in update_dict.items():
                data = user.update_json(key, value)

                resp = self.put_request(request_url, data)
                if "success" not in resp.text:
                    raise IDMRequestError(
                        f"Could not connect to IDM: Invalid combination\n{key=}\n{value=}\n{resp.text=}"
                    )
        else:
            data_list = []
            for key, value in update_dict.items():
                data_list.append(user.update_dict(key, value))
            base_data = data_list[0]
            for data_entry in data_list[1:]:
                base_data["attributes"].extend(data_entry["attributes"])
            resp = self.put_request(request_url, json.dumps(base_data))
            if "success" not in resp.text:
                raise IDMRequestError(
                    f"Could not connect to IDM: Invalid combination\n{data_list=}\n{resp.text=}"
                )
        return self.get_single(user.id)

    def create_new(self, new_user: CreateTemplate) -> ChangeTemplate:
        request_url: str = (
            f"{self.api_url}{self.api_suffix}/{new_user.create_template_name}"
        )
        data = new_user.to_json()
        resp = self.post_request(request_url, data)
        if "success" not in resp.text:
            raise IDMRequestError(
                f"Could not connect to IDM: Invalid combination\n{resp.text=}"
            )
        user_id = resp.headers["location"]
        return self.get_single(user_id)



from __future__ import annotations

from pydantic import Field

from ..attributes import SelfServiceAttributes

try:
    from ..base_models import SelfServiceBase
except ImportError:
    from ..core_models import IDMUserClass

    class SelfServiceBase(IDMUserClass):
        pass


from ..core_models import ChangeTemplate


class SelfService(SelfServiceBase):
    api_suffix: str = Field(default="/SelfService/objects")
    user_class: type[ChangeTemplate] = Field(default=SelfServiceAttributes)



from ..core_models import ChangeTemplate


class MailinglistenAttributes(ChangeTemplate):
    createTimestamp: str | None = None  # Creation date
    goesternMailRoutingAddresses: list[str] | None = None  # Routing addresses
    goesternProxyAddresses: list[str] | None = None  # Email addresses
    listName: str | None = None  # Listname
    modifyTimeStamp: str | None = None  # Modified on
    ou: str | None = None  # Institute



from .benutzerverwaltung import (  # noqa
    BenutzerverwaltungAttributes,
    BenutzerverwaltungAttributes_MDMP,
)
from .dynamischeexchangeverteiler import DynamischeExchangeVerteilerAttributes  # noqa
from .exchangeverteiler import ExchangeVerteilerAttributes  # noqa
from .gruppenverwaltung import (  # noqa
    GruppenverwaltungAttributes,
    GruppenverwaltungAttributes_MDMP,
)
from .ldapverteiler import LDAPVerteilerAttributes  # noqa
from .mailinglisten import MailinglistenAttributes  # noqa
from .ressourcen import RessourcenAttributes  # noqa
from .selfservice import SelfServiceAttributes  # noqa
from .sharedmailbox import SharedMailboxAttributes  # noqa
